/*
 * Copyright (C) 2002  Pekka Enberg <Pekka.Enberg@cs.helsinki.fi>
 *
 * Distributed under the terms of the GNU General Public License
 * version 2 or later.
 */
#include "xmlindent.h"
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define DEFAULT_INDENT_CHAR ' '
#define DEFAULT_NUM_INDENT_CHARS 4

char indent_char = DEFAULT_INDENT_CHAR;
int num_indent_chars = DEFAULT_NUM_INDENT_CHARS;
bool leave_comments_alone = false;

/* Print error message.  */
void error(const char * format, ...)
{
    va_list args;

    fprintf(stderr, "Error: ");

    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);

    fputc('\n', stderr);
}

/* Print indent characters.  */
void print_indent(FILE * output, int indent_level)
{
    int i;

    for (i = 0; i < (num_indent_chars * indent_level); i++) {
	fputc(indent_char, output);
    }
}

/* Buffer used to hold XML stream.  */
struct buffer {
    int * data;
    size_t length;
    size_t capacity;
};

/* Initialize buffer.  */
void buffer_init(struct buffer * buffer, size_t initial_capacity)
{
    buffer->data     = malloc(initial_capacity * sizeof(int));
    buffer->length   = 0;
    buffer->capacity = initial_capacity;
}

/* Release buffer.  */
void buffer_release(struct buffer * buffer)
{
    if (buffer->capacity > 0) {
	free(buffer->data);
	buffer->data     = NULL;
	buffer->capacity = 0;
    }
    buffer->length = 0;
}

/* Check if buffer is full.  */
bool buffer_is_full(struct buffer * buffer)
{
    return (buffer->capacity == buffer->length);
}

#define BUFFER_GROWTH_INCREMENT 256

/* Grow buffer by one.  */
void buffer_grow(struct buffer * buffer)
{
    buffer->data =
	realloc(buffer->data,
		(buffer->capacity + BUFFER_GROWTH_INCREMENT) * sizeof(int));

    if (buffer->data == NULL) {
	error("Could not allocate memory for buffer.");
	exit(1);
    }
    buffer->capacity += BUFFER_GROWTH_INCREMENT;
}

/* Push item at back of buffer.  */
void buffer_push(struct buffer * buffer, int item)
{
    if (buffer_is_full(buffer)) {
	buffer_grow(buffer);
    }
    buffer->data[buffer->length] = item;
    buffer->length++;
}

/* Pop item from back of buffer.  */
int buffer_pop(struct buffer * buffer)
{
    int result = -1;

    if (buffer->length > 0) {
	buffer->length--;
	result = buffer->data[buffer->length];
    } else {
	error("Buffer underflow.");
    }
    return result;
}

/* Flush buffer to output stream.  */
void buffer_flush(struct buffer * buffer, FILE * output)
{
    int i;

    for (i = 0; i < buffer->length; i++) {
	fputc(buffer->data[i], output);
    }
    buffer->length = 0;
}

struct parser_context {
    struct buffer * buffer;
    int lookahead;
    int current;
};

/* Check if for XML start-tag.  */
static inline bool is_xml_start_tag(struct parser_context * ctx)
{
    return (('<' == ctx->current) &&
	    (isalpha(ctx->lookahead) ||
	     ('_' == ctx->lookahead) ||
	     (':' == ctx->lookahead)));
}

/* Check for XML end-tag.  */
static inline bool is_xml_end_tag(struct parser_context * ctx)
{
    return ((('<' == ctx->current) && ('/' == ctx->lookahead)) ||
	    (('/' == ctx->current) && ('>' == ctx->lookahead)));
}

/* Check if input matches token.
   Note: Only three first characters are actually checked.  */
static inline bool match(struct parser_context * ctx, FILE * input,
			 const char * token)
{
    bool result = false;

    if ((token[0] == ctx->current) && (token[1] == ctx->lookahead)) {
	int tmp = fgetc(input);
	if (token[2] == tmp) {
	    result = true;
	}
	if (ungetc(tmp, input) != tmp) {
	    error("Could not restore XML stream.");
	}
    }
    return result;
}

/* Check if input matches token.  */
static inline bool match_backtrack(struct parser_context * ctx,
				   const char * token)
{
    bool result = false;

    if ((token[strlen(token) - 2] == ctx->current) &&
	(token[strlen(token) - 1] == ctx->lookahead)) {
	int tmp = buffer_pop(ctx->buffer);

	if (token[strlen(token) - 3] == tmp) {
	    result = true;
	}
	buffer_push(ctx->buffer, tmp);
    }
    return result;
}

/* Check for XML comment start.  */
static inline bool is_xml_comment_start(struct parser_context * ctx,
					FILE * input)
{
    return match(ctx, input, "<!--");
}

/* Check for XML comment end.  */
static inline bool is_xml_comment_end(struct parser_context * ctx)
{
    return match_backtrack(ctx, "-->");
}

/* Check for CDATA section start.  */
static inline bool is_cdata_section_start(struct parser_context * ctx,
					  FILE * input)
{
    return match(ctx, input, "<![CDATA[");
}

/* Check for CDATA section end.  */
static inline bool is_cdata_section_end(struct parser_context * ctx)
{
    return match_backtrack(ctx, "]]>");
}

/* XML end of line characters.  */
#define CARRIAGE_RETURN 0x0D
#define LINE_FEED       0x0A
#define NEL             0x85

static inline bool is_end_of_line(struct parser_context * ctx)
{
    if ((CARRIAGE_RETURN == ctx->current) ||
	(LINE_FEED == ctx->current) ||
	((CARRIAGE_RETURN == ctx->current) && (LINE_FEED == ctx->lookahead)) ||
	(NEL == ctx->current) ||
	((CARRIAGE_RETURN == ctx->current) && (NEL == ctx->lookahead)))
	return true;

    return false;
}

static inline bool is_two_char_end_of_line_marker(struct parser_context * ctx)
{
    if (((CARRIAGE_RETURN == ctx->current) && (LINE_FEED == ctx->lookahead)) ||
	((CARRIAGE_RETURN == ctx->current) && (NEL == ctx->lookahead)))
	return true;

    return false;
}

/* Check for whitespace character.  */
static inline bool is_whitespace(int c)
{
    return ((c == ' ') || (c == '\f') || (c == '\t') || (c == '\v'));
}

/* Eat whitespace from stream.  */
void eat_whitespace(FILE * input)
{
    for (;;) {
	int current = fgetc(input);
	if ((is_whitespace(current) == 0)) {
	    if (ungetc(current, input) != current) {
		error("Could not restore XML stream.");
	    }
	    break;
	}
    }
}

/* Pushes CDATA section to buffer.  */
void push_cdata_section_to_buffer(struct parser_context * ctx, FILE * input)
{
    buffer_push(ctx->buffer, ctx->current);

    for (;;) {
	ctx->current = ctx->lookahead;

	if (EOF == ctx->current) {
	    error("Premature end of file.");
	    break;
	}
	ctx->lookahead = fgetc(input);

	if (is_cdata_section_end(ctx)) {
	    break;
	}
	buffer_push(ctx->buffer, ctx->current);
    }
}

#define BUFFER_INITIAL_CAPACITY 1024

/* Indent XML stream.  */
void indent(FILE * input, FILE * output)
{
    struct parser_context parser_context;
    struct parser_context * ctx = &parser_context;
    struct buffer buffer;

    int indent_level;
    int start_tag_count;
    bool inside_comment;

    ctx->buffer    = &buffer;
    ctx->lookahead = fgetc(input);

    indent_level         = 0;
    start_tag_count      = 0;
    inside_comment       = false;

    buffer_init(ctx->buffer, BUFFER_INITIAL_CAPACITY);

    for (;;) {
	ctx->current = ctx->lookahead;

	if (EOF == ctx->current) {
	    break;
	}
	ctx->lookahead = fgetc(input);

	if (!inside_comment) {
	    if (is_xml_start_tag(ctx)) {
		/* XML start-tag changes indent level only after we have
		   finished the current line.  */
		start_tag_count++;
	    } else if (is_xml_end_tag(ctx)) {
		if (start_tag_count > 0) {
		    /* start_tag_count keeps track of how many
                       start-tags are on the same line.  */
		    start_tag_count--;
		} else {
		    indent_level--;
		    if (indent_level < 0) {
			error("XML start-tags and end-tags are not balanced.");
			indent_level = 0;
		    }
		}
	    } else if (is_xml_comment_start(ctx, input)) {
		inside_comment = true;
	    } else if (is_cdata_section_start(ctx, input)) {
		push_cdata_section_to_buffer(ctx, input);
	    }
	} else if (is_xml_comment_end(ctx)) {
	    inside_comment = false;
	}

	buffer_push(ctx->buffer, ctx->current);

	if (is_end_of_line(ctx)) {
	    if (is_two_char_end_of_line_marker(ctx)) {
		buffer_push(ctx->buffer, ctx->lookahead);
		/* Eat lookahead.  */
		ctx->lookahead = fgetc(input);
	    }

	    if (inside_comment && leave_comments_alone) {
		continue;
	    }
	    print_indent(output, indent_level);
	    buffer_flush(ctx->buffer, output);

	    if (start_tag_count > 0) {
		indent_level += start_tag_count;
		start_tag_count = 0;
	    }

	    if (ungetc(ctx->lookahead, input) != ctx->lookahead) {
		error("Could not restore XML stream.");
	    }
	    eat_whitespace(input);
	    ctx->lookahead = fgetc(input);
	}
    }

    buffer_release(ctx->buffer);
}
